"""
Support functions for JADN codec
  Convert dict between nested and flat
  Convert typedef options between dict and strings
"""

import re
from functools import reduce
import jadn
from jadn.definitions import *


# Dict conversion utilities


def _dmerge(x, y):
    k, v = next(iter(y.items()))
    if k in x:
        _dmerge(x[k], v)
    else:
        x[k] = v
    return x


def hdict(keys, value, sep='.'):
    """
    Convert a hierarchical-key value pair to a nested dict
    """
    return reduce(lambda v, k: {k: v}, reversed(keys.split(sep)), value)


def fluff(src, sep='.'):
    """
    Convert a flat dict with hierarchical keys to a nested dict

    :param src: flat dict - e.g., {'a.b.c': 1, 'a.b.d': 2}
    :param sep: separator character for keys
    :return: nested dict - e.g., {'a': {'b': {'c': 1, 'd': 2}}}
    """
    return reduce(lambda x, y: _dmerge(x, y), [hdict(k, v, sep) for k, v in src.items()], {})


def flatten(cmd, path='', fc=None, sep='.'):
    """
    Convert a nested dict to a flat dict with hierarchical keys
    """
    if fc is None:
        fc = {}
    fcmd = fc.copy()
    if isinstance(cmd, dict):
        for k, v in cmd.items():
            k = k.split(':')[1] if ':' in k else k
            fcmd = flatten(v, sep.join((path, k)) if path else k, fcmd)
    elif isinstance(cmd, list):
        for n, v in enumerate(cmd):
            fcmd.update(flatten(v, sep.join([path, str(n)])))
    else:
        fcmd[path] = cmd
    return fcmd


def dlist(src):
    """
    Convert dicts with numeric keys to lists

    :param src: {'a': {'b': {'0':'red', '1':'blue'}, 'c': 'foo'}}
    :return: {'a': {'b': ['red', 'blue'], 'c': 'foo'}}
    """
    if isinstance(src, dict):
        for k in src:
            src[k] = dlist(src[k])
        if set(src) == set([str(k) for k in range(len(src))]):
            src = [src[str(k)] for k in range(len(src))]
    return src


def build_deps(schema):         # Build a Dependency dict: {TypeName: {Dep1, Dep2, ...}}
    def get_refs(tdef):         # Return all type references from a type definition
        oids = [OPTION_ID['ktype'], OPTION_ID['vtype'], OPTION_ID['and'], OPTION_ID['or']]  # Options whose value is/has a type name
        oids2 = [OPTION_ID['enum'], OPTION_ID['pointer']]                       # Options that enumerate fields
        refs = [to[1:] for to in tdef[TypeOptions] if to[0] in oids and not is_builtin(to[1:])]
        refs += [to for to in tdef[TypeOptions] if to[0] in oids2]
        if has_fields(tdef[BaseType]):
            for f in tdef[Fields]:
                if not is_builtin(f[FieldType]):
                    refs.append(f[FieldType])                               # Add reference to type name
                refs += get_refs(['', f[FieldType], f[FieldOptions], ''])   # Get refs from type opts in field (extension)
        return refs

    return {t[TypeName]: get_refs(t) for t in schema['types']}


def topo_sort(items):
    """
    Topological sort with locality
    Sorts a list of (item: (dependencies)) pairs so that 1) all dependency items are listed before the parent item,
    and 2) dependencies are listed in the given order and as close to the parent as possible.
    Returns the sorted list of items and a list of root items.  A single root indicates a fully-connected hierarchy;
    multiple roots indicate disconnected items or hierarchies, and no roots indicate a dependency cycle.
    """
    def walk_tree(it):
        for i in deps[it]:
            if i not in out:
                walk_tree(i)
                out.append(i)

    out = []
    deps = {i[0]: i[1] for i in items}           # TODO: update for items dict
    roots = {i[0] for i in items} - set().union(*[i[1] for i in items])
    for item in roots:
        walk_tree(item)
        out.append(item)
    out = out if out else [i[0] for i in items]     # if cycle detected, don't sort
    return out, roots


def get_optx(opts, oname):
    n = [i for i, x in enumerate(opts) if x[0] == OPTION_ID[oname]]
    return n[0] if n else None


def del_opt(opts, oname):
    n = [i for i, x in enumerate(opts) if x[0] == OPTION_ID[oname]]
    if n:
        del opts[n[0]]


def topts_s2d(olist, frange=False):        # Convert list of type definition option strings to options dictionary
    assert isinstance(olist, (list, tuple)), f'{olist} is not a list'
    opts = {}
    for o in olist:
        try:
            k = TYPE_OPTIONS[ord(o[0])]
            opts[k[0]] = k[1](o[1:])
        except KeyError:
            jadn.raise_error(f'Unknown type option: {o}')
    return opts


def ftopts_s2d(olist):       # Convert list of field definition option strings to options dictionary
    assert isinstance(olist, (list, tuple)), f'{olist} is not a list'
    fopts, topts = {}, {}
    for o in olist:
        try:
            k = FIELD_OPTIONS[ord(o[0])]
            fopts[k[0]] = k[1](o[1:])
        except KeyError:
            topts.update(topts_s2d([o]))
    return fopts, topts


def opts_d2s(to):
    return [OPTION_ID[k] + ('' if v is True else str(v)) for k, v in to.items()]


def opts_sort(olist):      # Sort JADN option list into canonical order
    def opt_order(o):
        try:
            k = FIELD_OPTIONS[ord(o)][2]
        except KeyError:
            k = TYPE_OPTIONS[ord(o)][2]
        return k

    olist.sort(key=lambda x: opt_order(x[0]))


def canonicalize(schema):
    """
    """
    def can_opts(olist):
        opts_sort(olist)        # Sort options into canonical order (for comparisons)
                                # TODO: remove default min-size and multiplicity options
                                # y/z options at least one decimal

    for td in schema['types']:
        can_opts(td[TypeOptions])
        for fd in td[Fields]:
            if td[BaseType] != 'Enumerated':
                can_opts(fd[FieldOptions])
    return schema


def cleanup_tagid(fields):          # If type definition contains a TagId option, replace field name with id
    for f in fields:
        if len(f) > FieldOptions:
            tx = get_optx(f[FieldOptions], 'tagid')
            if tx is not None:
                to = f[FieldOptions][tx]
                try:
                    tag = int(to[1:])           # Check if already a Field Id
                except ValueError:              # Look up Id corresponding to Field Name
                    fx = {x[FieldName]: x[FieldID] for x in fields}
                    f[FieldOptions][tx] = to[0] + str(fx[to[1:]])
    return fields


def typestr2jadn(typestring):
    def parseopt(optstr):
        mm = re.match(r'^\s*([-$\w]+)(?:\[([^]]+)\])?$', optstr)
        return OPTION_ID[mm.group(1).lower()] + mm.group(2) if mm.group(2) else mm.group(1)

    topts = {}
    fo = []
    p_name = r'\s*=?\s*([-$:\w]+)'      # 1 type name
    p_id = r'(.ID)?'                    # 2 'id'
    p_func = r'(?:\(([^)]+)\))?'        # 3 'ktype', 'vtype', 'enum', 'pointer', 'tagid'
    p_range = r'(?:\s*\{(.*)\})?'       # 4 'minv', 'maxv'
    p_format = r'(?:\s*\/(\w[-\w]*))?'  # 5 'format'
    p_pattern = r'(?:\s*\(%(.+)%\))?'   # 6 'pattern'
    p_unique = r'\s*(unique)?'          # 7 'unique'
    pattern = '^' + p_name + p_id + p_func + p_range + p_format + p_pattern + p_unique + '\s*\{?\s*$'
    m = re.match(pattern, typestring)
    if m is None:
        return '*ERR*', [], []
    tname = m.group(1)
    topts.update({'id': True} if m.group(2) else {})
    if m.group(3):                      # (ktype, vtype), Enum(), Pointer() options
        opts = [parseopt(x) for x in m.group(3).split(',', maxsplit=1)]
        if tname == 'MapOf':
            topts.update({'ktype': opts[0], 'vtype': opts[1]})
        elif tname == 'ArrayOf':
            assert len(opts) == 1
            topts.update({'vtype': opts[0]})
        else:
            assert len(opts) == 1
            topts.update(topts_s2d([opts[0]]) if ord(opts[0][0]) in TYPE_OPTIONS else {})
            fo += [opts[0]] if ord(opts[0][0]) in FIELD_OPTIONS else []         # TagId option
    if m.group(4):
        a, b = m.group(4).split('..', maxsplit=1)
        if tname == 'Number':
            topts.update({} if a == '*' else {'minf': float(a)})
            topts.update({} if b == '*' else {'maxf': float(b)})
        else:
            a = '*' if tname != 'Integer' and a != '*' and int(a) == 0 else a   # Default min size = 0
            topts.update({} if a == '*' else {'minv': int(a)})
            topts.update({} if b == '*' else {'maxv': int(b)})
    topts.update({'format': m.group(5)} if m.group(5) else {})
    topts.update({'pattern': m.group(6)} if m.group(6) else {})
    topts.update({'unique': True} if m.group(7) else {})
    return tname, opts_d2s(topts), fo


def jadn2typestr(tname, topts):     # Convert typename and options to string

    def _kvstr(optv):               # Handle ktype/vtype containing Enum options
        if optv[0] == OPTION_ID['enum']:
            return 'Enum[' + optv[1:] + ']'
        elif optv[0] == OPTION_ID['pointer']:
            return 'Pointer[' + optv[1:] + ']'
        return optv

    def _srange(ops):               # Size range (single-ended) - default is {0..*}
        lo = ops.pop('minv', 0)
        hi = ops.pop('maxv', 0)
        hs = '*' if hi == 0 else str(hi)
        return str(lo) + '..' + hs if lo != 0 or hi != 0 else ''

    def _vrange(ops):               # Value range (double-ended) - default is {*..*}
        lo = ops.pop('minv', '*')
        hi = ops.pop('maxv', '*')
        return str(lo) + '..' + str(hi) if lo != '*' or hi != '*' else ''

    def _frange(ops):               # Value range (double-ended) - default is {*..*}
        lo = ops.pop('minf', '*')
        hi = ops.pop('maxf', '*')
        return str(lo) + '..' + str(hi) if lo != '*' or hi != '*' else ''

    opts = topts_s2d(topts)
    extra = '.ID' if opts.pop('id', None) else ''   # SIDE EFFECT!: remove known options from opts.
    if tname == 'ArrayOf':
        extra += '(' + _kvstr(opts.pop('vtype')) + ')'
    elif tname == 'MapOf':
        extra += '(' + _kvstr(opts.pop('ktype')) + ', ' + _kvstr(opts.pop('vtype')) + ')'
    v = opts.pop('enum', None)
    extra += '(Enum[' + v + '])' if v else ''
    v = opts.pop('pointer', None)
    extra += '(Pointer[' + v + '])' if v else ''
    v = opts.pop('pattern', None)
    extra += ' (%' + v + '%)' if v else ''
    v = _vrange(opts) if tname == 'Integer' else (_frange(opts) if tname == 'Number' else _srange(opts))
    extra += '{' + v + '}' if v else ''
    v = opts.pop('format', None)
    extra += ' /' + v if v else ''
    v = opts.pop('unique', None)
    extra += ' unique' if v else ''
    v = opts.pop('and', None)           # hack set operations for now.  TODO: generalize to any number
    extra += ' ∩ ' + v if v else ''
    v = opts.pop('or', None)
    extra += ' ∪ ' + v if v else ''
    return tname + extra + (' ?' + str([str(k) for k in opts]) + '?' if opts else '')  # Flag unrecognized options


def jadn2fielddef(fdef, tdef):
    idt = tdef[BaseType] == 'Array' or get_optx(tdef[TypeOptions], 'id') is not None
    fname = '' if idt else fdef[FieldName]
    ftyperef, fmult = '', ''
    fdesc = fdef[FieldName] + ':: ' if idt else ''
    if tdef[BaseType] == 'Enumerated':
        fdesc += fdef[ItemDesc]
    else:
        fdesc += fdef[FieldDesc]
        fo, fto = ftopts_s2d(fdef[FieldOptions])
        fname += '/' if 'dir' in fo else ''
        tagid = fo.get('tagid')
        tf = ''
        if tagid:
            tf = {f[FieldID]: f[FieldName] for f in tdef[Fields]}[tagid]
            tf = tf if tf else str(tagid)
            tf = '(TagId[' + tf + '])'
        ftyperef = jadn2typestr(fdef[FieldType] + tf, opts_d2s(fto))
        minc, maxc = fo.get('minc', 1), fo.get('maxc', 1)
        fmult = '1' if minc == 1 and maxc == 1 else str(minc) + '..' + ('*' if maxc == 0 else str(maxc))
    return fname, ftyperef, fmult, fdesc


def fielddef2jadn(fid, fname, fstr, fmult, fdesc):
    ftyperef = ''
    fo = {}
    if fstr:
        ftyperef, topts, fopts = typestr2jadn(fstr)
        # Field is one of: enum.id, enum, field.id, field
        fo = topts_s2d(topts)                   # Copy type options (if any) into field options (JADN extension)
        if fname.endswith('/'):
            fo.update({'dir': True})
            fname = fname.rstrip('/')
        m = re.match(r'^(\d+)\.\.(\d+|\*)$', fmult) if fmult else None
        if m:
            minc = int(m.group(1))
            maxc = 0 if m.group(2) == '*' else int(m.group(2))
            fo.update({'minc': minc} if minc != 1 else {})
            fo.update({'maxc': maxc} if maxc != 1 else {})
        if fopts:
            assert len(fopts) == 1 and fopts[0][0] == OPTION_ID['tagid']     # Update if additional field options defined
            fo.update({'tagid': fopts[0][1:]})      # if field name, MUST update to id after all fields have been read
    if fdesc:
        m = re.match(r'^(?:\s*\/\/)?\s*(.*)$', fdesc)
        fdesc = m.group(1)
        if not fname:
            m = re.match(r'^([^:]+)::\s*(.*)$', fdesc)
            if m:
                fname = m.group(1)
                fdesc = m.group(2)
    return [fid, fname, ftyperef, opts_d2s(fo), fdesc] if ftyperef else [fid, fname, fdesc]


def get_config(meta):
    config = dict(DEFAULT_CONFIG)
    config.update(meta['config'] if meta and 'config' in meta else {})
    return config



import re
import json
from datetime import datetime
from lxml import etree, html
from lxml.html.builder import *
from jadn.definitions import *
from jadn import raise_error
from jadn import jadn2typestr, typestr2jadn, jadn2fielddef, fielddef2jadn, cleanup_tagid, get_optx


"""
Convert JADN schema to and from HTML format

Generated HTML uses "div tables" rather than HTML table elements to support responsive page design.
Css styles for classes beginning with 't' must be defined to render correctly.

Generated HTML includes JADN Schema classes to support parsing.
Css styles for classes beginning with 'j' are not required

CSS class definitions needed to render table layout:
tTable:   display table
tCaption: display table-caption
tHead:    display table-header-group
tHCell:   display table-cell (used in header)
tBody:    display table-row-group
tRow:     display table-row
tCell:    display table-cell (used in body)

JADN schema elements (required for parsing, available for styling):
jKey:     metadata key
jVal:     metadata value
jTname:   TypeName - name of a defined type
jTstr:    String representation of BaseType and TypeOptions
jTdesc:   TypeDesc - description of a type
jFid:     FieldId - integer tag of a field or enumerated item
jFname:   FieldName/ItemValue - name of a field or string value of an enumerated item
jFstr:    String representation of FieldType and FieldOptions (except multiplicity)
jFmult:   String representation of multiplicity FieldOptions
jFdesc:   FieldDesc/ItemDesc - description of a field in a container type or item in an enumerated type

JADN styling classes (not used for parsing):
jMeta:    metadata section
jType:    type definition
jHid:     FieldId
jHname:   FieldName
jHstr:    FieldType + FieldOptions
jHmult:   Multiplicity
jHdesc:   Description
"""

"""
Convert JADN to HTML
"""

def html_dumps(schema):
    try:
        title = schema['meta']['title']
    except KeyError:
        title = 'JADN Schema'
    doc = HTML(ATTR({'lang': 'en'}),        # Make initial etree
        HEAD(
            META(charset='UTF-8'),
            TITLE(title),
            LINK(rel='stylesheet', href='dtheme.css', type='text/css')
        ),
        BODY(
            H2('Schema')
        )
    )
    body = doc.find('body')

    if 'meta' in schema:                    # Add meta elements if present
        het = etree.Element('div', {'class': 'tTable jMeta'})       # top-level element of the metadata table
        he2 = etree.SubElement(het, 'div', {'class': 'tBody'})
        mlist = [k for k in META_ORDER if k in schema['meta']]
        for k in mlist + list(set(schema['meta']) - set(mlist)):
            he3 = etree.SubElement(he2, 'div', {'class': 'tRow'})
            etree.SubElement(he3, 'div', {'class': 'tCell jKey'}).text = k + ':'
            etree.SubElement(he3, 'div', {'class': 'tCell jVal'}).text = json.dumps(schema['meta'][k])
        body.append(het)

    for tdef in schema['types']:            # Add type definitions
        het = etree.Element('div', {'class': 'tTable jType'})       # top-level element of a type definition
        he2 = etree.SubElement(het, 'div', {'class': 'tCaption'})
        he3 = etree.SubElement(he2, 'div', {'class': 'jTdef'})      # container for type definition column
        etree.SubElement(he3, 'div', {'class': 'jTname'}).text = tdef[TypeName]
        etree.SubElement(he3, 'div', {'class': 'jTstr'}).text = ' = ' + jadn2typestr(tdef[BaseType], tdef[TypeOptions])
        etree.SubElement(he2, 'div', {'class': 'jTdesc'}).text = ' // ' + tdef[TypeDesc] if tdef[TypeDesc] else ''
        if len(tdef) > Fields:
            he2 = etree.SubElement(het, 'div', {'class': 'tHead'})
            etree.SubElement(he2, 'div', {'class': 'tHCell jHid'}).text = 'ID'
            etree.SubElement(he2, 'div', {'class': 'tHCell jHname'}).text = 'Name'
            if tdef[Fields] and len(tdef[Fields][0]) > ItemDesc + 1:
                etree.SubElement(he2, 'div', {'class': 'tHCell jHstr'}).text = 'Type'
                etree.SubElement(he2, 'div', {'class': 'tHCell jHmult'}).text = '#'
            etree.SubElement(he2, 'div', {'class': 'tHCell jHdesc'}).text = 'Description'
            he2 = etree.SubElement(het, 'div', {'class': 'tBody'})
            for fdef in tdef[Fields]:
                he3 = etree.SubElement(he2, 'div', {'class': 'tRow'})
                if len(tdef[Fields][0]) > ItemDesc + 1:
                    etree.SubElement(he3, 'div', {'class': 'tCell jFid'}).text = str(fdef[FieldID])
                    fname, ftyperef, fmult, fdesc = jadn2fielddef(fdef, tdef)
                    etree.SubElement(he3, 'div', {'class': 'tCell jFname'}).text = fname
                    etree.SubElement(he3, 'div', {'class': 'tCell jFstr'}).text = ftyperef
                    etree.SubElement(he3, 'div', {'class': 'tCell jFmult'}).text = fmult
                    etree.SubElement(he3, 'div', {'class': 'tCell jFdesc'}).text = fdesc
                else:
                    etree.SubElement(he3, 'div', {'class': 'tCell jFid'}).text = str(fdef[ItemID])
                    etree.SubElement(he3, 'div', {'class': 'tCell jFname'}).text = fdef[ItemValue]  # TODO: id option
                    etree.SubElement(he3, 'div', {'class': 'tCell jFdesc'}).text = fdef[ItemDesc]
        body.append(het)

    return html.tostring(doc, pretty_print=True, doctype='<!DOCTYPE html>').decode('utf-8')


def html_dump(schema, fname, source=''):
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write('<! Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + '>\n\n')
        f.write(html_dumps(schema))


"""
Convert HTML to JADN
"""


def line2jadn(lt, tdef):
    def default(x, d):
        return x if x else d

    if 'jKey' in lt:
        return 'M', (lt['jKey'].rstrip(':'), lt['jVal'])
    if 'jTname' in lt:
        btype, topts, fo = typestr2jadn(lt['jTstr'])
        assert fo == []                     # field options MUST not be included in typedefs
        tdesc = default(lt.get('jTdesc', ''), '')
        if tdesc:
            m = re.match(r'^(?:\s*\/\/)?\s*(.*)$', tdesc)
            tdesc = m.group(1)
        return 'T', [lt['jTname'], btype, topts, tdesc, []]
    if 'jFid' in lt:
        fname = default(lt.get('jFname', ''), '')
        fstr = default(lt.get('jFstr', ''), '')
        fdesc = default(lt.get('jFdesc', ''), '')
        fmult = lt.get('jFmult', '')
        return 'F', fielddef2jadn(int(lt['jFid']), fname, fstr, fmult, fdesc)


def html_loads(hdoc):
    def get_line(hdoc):
        line = {}
        tclass = ('jKey', 'jVal', 'jTname', 'jTstr', 'jTdesc', 'jFid', 'jFname', 'jFstr', 'jFmult', 'jFdesc')
        for element in html.fromstring(hdoc).iter():
            cl = [c for c in element.get('class', '').split() if c in tclass]    # Get element's token class
            if cl:
                assert len(cl) == 1     # TODO: Can't be more than one - replace assertions with proper errors
                if cl[0] in ('jKey', 'jTname', 'jFid'):
                    if line:
                        yield line
                        line = {}
                line.update({cl[0]: element.text})
        yield line

    meta = {}
    types = []
    fields = None
    for n, line in enumerate(get_line(hdoc), start=1):
        if line:
            t, v = line2jadn(line, types[-1] if types else None)
            if t == 'F':
                fields.append(v)
            elif fields:
                cleanup_tagid(fields)
                fields = None
            if t == 'M':
                meta.update({v[0]: json.loads(v[1])})
            elif t == 'T':
                types.append(v)
                fields = types[-1][Fields]
    return {'meta': meta, 'types': types} if meta else {'types': types}


def html_load(fname):
    with open(fname, 'r') as f:
        return html_loads(f.read())



"""
Translate JADN to JADN Interface Definition Language
"""
import json
from datetime import datetime
import re
from jadn.definitions import *
from jadn import topts_s2d, ftopts_s2d, opts_d2s, get_optx, raise_error
from jadn import jadn2typestr, typestr2jadn, jadn2fielddef, fielddef2jadn, cleanup_tagid


"""
Convert JADN to JIDL
"""


def jidl_dumps(schema, width=None):
    w = {
        'meta': 12,
        'id': 4,
        'name': 12,
        'type': 35,
        'page': None
    }
    if width:
        w.update(width)

    text = ''
    meta = schema['meta'] if 'meta' in schema else {}
    mlist = [k for k in META_ORDER if k in meta]
    for k in mlist + list(set(meta) - set(mlist)):              # Display meta elements in fixed order
        text += f'{k:>{w["meta"]}}: {json.dumps(meta[k])}\n'    # TODO: wrap to page width, continuation-line parser

    wt = w['id'] + w['name'] + w['type']
    for td in schema['types']:
        tdef = f'{td[TypeName]} = {jadn2typestr(td[BaseType], td[TypeOptions])}'
        tdesc = '// ' + td[TypeDesc] if td[TypeDesc] else ''
        text += f'\n{tdef:<{wt}}{tdesc}'[:w['page']] + '\n'
        idt = td[BaseType] == 'Array' or get_optx(td[TypeOptions], 'id') is not None
        for fd in td[Fields] if len(td) > Fields else []:       # TODO: constant-length types
            fname, fdef, fmult, fdesc = jadn2fielddef(fd, td)
            if td[BaseType] == 'Enumerated':
                fdesc = '// ' + fdesc if fdesc else ''
                fs = f'{fd[ItemID]:>{w["id"]}} {fname}'
                wf = w['id'] + w['name'] + 2
            else:
                fdef += '' if fmult == '1' else ' optional' if fmult == '0..1' else ' [' + fmult + ']'
                fdesc = '// ' + fdesc if fdesc else ''
                wn = 0 if idt else w['name']
                fs = f'{fd[FieldID]:>{w["id"]}} {fname:<{wn}} {fdef}'
                wf = w['id'] + w['type'] if idt else wt
            text += f'{fs:{wf}}{fdesc}'[:w['page']] + '\n'
    return text


def jidl_dump(schema, fname, source=''):
    with open(fname, 'w', encoding='utf8') as f:
        if source:
            f.write('/* Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + ' */\n\n')
        f.write(jidl_dumps(schema))


"""
Convert JIDL to JADN
"""


def line2jadn(line, tdef):
    if line:
        p_meta = r'^\s*([-\w]+):\s*(.+?)\s*$'
        m = re.match(p_meta, line)
        if m:
            return 'M', (m.group(1), m.group(2))

        p_tname = r'\s*([-$\w]+)'               # Type Name
        p_assign = r'\s*='                      # Type assignment operator
        p_tstr  = r'\s*(.*?)\s*\{?'             # Type definition
        p_tdesc = r'(?:\s*\/\/\s*(.*?)\s*)?'    # Optional Type description
        p_type = '^' + p_tname + p_assign + p_tstr + p_tdesc + '$'
        m = re.match(p_type, line)
        if m:
            btype, topts, fo = typestr2jadn(m.group(2))
            assert fo == []                     # field options MUST not be included in typedefs
            newtype = [m.group(1), btype, topts, m.group(3) if m.group(3) else '', []]
            return 'T', newtype

        p_id = r'\s*(\d+)'                      # Field ID
        p_fname = r'\s+([-:$\w]+\/?)?'          # Field Name with dir/ option (colon is deprecated, allow for now)
        p_fstr = r'\s*(.*?),?'                  # Field definition or Enum value
        p_range = r'\s*(?:\[([.*\d]+)\]|(optional))?,?'     # Multiplicity
        p_desc = r'\s*(?:\/\/\s*(.*?)\s*)?'     # Field description, including field name if .id option
        pn = '()' if (get_optx(tdef[TypeOptions], 'id') is not None or tdef[BaseType] == 'Array') else p_fname
        if tdef[BaseType] == 'Enumerated':      # Parse Enumerated Item
            pattern = '^' + p_id + p_fstr + p_desc + '$'
            m = re.match(pattern, line)
            if m:
                return 'F', fielddef2jadn(int(m.group(1)), m.group(2), '', '', m.group(3) if m.group(3) else '')
        else:                                   # Parse Field
            pattern = '^' + p_id + pn + p_fstr + p_range + p_desc + '$'
            m = re.match(pattern, line)
            if m:
                range = '0..1' if m.group(5) else m.group(4)        # Convert 'optional' to range
                fdesc = m.group(6) if m.group(6) else ''
                return 'F', fielddef2jadn(int(m.group(1)), m.group(2), m.group(3), range if range else '', fdesc)

        if line.strip() not in ('', '}'):
            raise_error(f'JIDL load{repr(line)}')
    return '', ''


def jidl_loads(doc):
    meta = {}
    types = []
    fields = None
    for n, line in enumerate(doc.splitlines(), start=1):
        if line:
            t, v = line2jadn(line, types[-1] if types else None)    # Parse a JIDL line
            if t == 'F':
                fields.append(v)
            elif fields:
                cleanup_tagid(fields)
                fields = None
            if t == 'M':
                meta.update({v[0]: json.loads(v[1])})
            elif t == 'T':
                types.append(v)
                fields = types[-1][Fields]
    return {'meta': meta, 'types': types} if meta else {'types': types}


def jidl_load(fname):
    with open(fname, 'r') as f:
        return jidl_loads(f.read())


"""
Load, validate, prettyprint, and dump JSON Abstract Encoding Notation (JADN) schemas
"""

import json
import jsonschema
import numbers
import os
from datetime import datetime

import jadn
from jadn.definitions import *


def data_dir():
    """
    Return directory containing JADN schema files
    """
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), 'data')


def raise_error(*s):                     # Handle errors
    raise ValueError(*s)


def check(schema):
    """
    Validate JADN schema against JSON schema,
    Validate JADN schema against JADN meta-schema, then
    Perform additional checks on type definitions
    """

    def check_typeopts(type_name, base_type, topts):  # Check for invalid type options and undefined formats
        ro = {k for k in REQUIRED_TYPE_OPTIONS[base_type]} - {k for k in topts}
        if ro:
            raise_error(f'Missing type option {type_name}: {str(ro)}')
        uo = {k for k in topts} - {k for k in ALLOWED_TYPE_OPTIONS[base_type]}
        if uo:
            raise_error(f'Unsupported type option {type_name} ({base_type}): {str(uo)}')
        if 'format' in topts:
            f = topts['format']
            fm = dict(list(FORMAT_VALIDATE.items()) + list(FORMAT_JS_VALIDATE.items()) + list(FORMAT_SERIALIZE.items()))
            if f not in fm or base_type != fm[f]:
                raise_error(f'Unsupported format {f} in {type_name} {base_type}')
        if 'enum' in topts and 'pointer' in topts:
            raise_error(f'Type cannot be both Enum and Pointer {type_name} {base_type}')
        if 'and' in topts and 'or' in topts:
            raise_error(f'Unsupported union+intersection in {type_name} {base_type}')

    for t in schema['types']:           # Transition type definitions to constant length
        if len(t) <= Fields:
            t.append([])

    here = data_dir()
    with open(os.path.join(here, 'jadn_v1.0_schema.json')) as f:     # Check using JSON Schema for JADN
        jsonschema.Draft7Validator(json.load(f)).validate(schema)

    with open(os.path.join(here, 'jadn_v1.0_schema.jadn')) as f:     # Optional: check using JADN meta-schema
        meta_schema = jadn.codec.Codec(json.load(f), verbose_rec=True, verbose_str=True, config=schema)
        assert meta_schema.encode('Schema', schema) == schema

    types = set()                                               # Additional checks not included in schema
    for t in schema['types']:
        tn = t[TypeName]
        bt = t[BaseType]
        if tn in types:
            raise_error(f'Duplicate type definition {tn}')
        types |= {tn}
        if is_builtin(tn):
            raise_error(f'Reserved type name {tn}')
        if not is_builtin(bt):
            raise_error(f'Invalid base type {tn}: {bt}')
        topts = jadn.topts_s2d(t[TypeOptions])
        check_typeopts(tn, bt, topts)
        flen = 0 if 'enum' in topts or 'pointer' in topts else FIELD_LENGTH[bt]
        if flen:                # Check fields
            fids = set()        # Field IDs
            fnames = set()      # Field Names
            ordinal = bt in ('Array', 'Record')
            for n, f in enumerate(t[Fields]):
                if len(f) != flen:
                    raise_error(f'Bad field {n+1} in {tn} length, {len(f)} should be {flen}')
                fids.update({f[FieldID]})
                fnames.update({f[FieldName]})
                if ordinal and f[FieldID] != n + 1:
                    raise_error(f'Item tag error: {tn}({bt}) [{f[FieldName]}] -- {f[FieldID]} should be {n+1}')
                if flen > FieldDesc:    # Full field, not an Enumerated item
                    fo, fto = jadn.ftopts_s2d(f[FieldOptions])
                    minc, maxc = fo.get('minc', 1), fo.get('maxc', 1)
                    if minc < 0 or maxc < 0 or (maxc > 0 and maxc < minc):
                        raise_error(f'{tn}/{f[FieldName]} bad cardinality {minc} {maxc}')
                    tf = fo.get('tagid')
                    if tf and tf not in fids:
                        raise_error(f'{tn}/{f[FieldName]}({f[FieldType]}) choice has bad external tag {tf}')
                    if is_builtin(f[FieldType]):
                        check_typeopts(f'{tn}/{f[FieldName]}', f[FieldType], fto)
                    elif fto:
                        allowed = {'unique', } if maxc != 1 else set()  # unique option is moved to generated ArrayOf
                        if set(fto) - allowed:
                            raise_error(f'{tn}/{f[FieldName]}({f[FieldType]}) cannot have Type options {fto}')

            if len(t[Fields]) != len(fids) or len(t[Fields]) != len(fnames):
                raise_error(f'Duplicate field {tn} {len(t[Fields])} fields, {len(fids)} unique tags, {len(fnames)} unique names')
    return schema


def analyze(schema):
    def ns(name, nsids):   # Return namespace if name has a known namespace, otherwise return full name
        nsp = name.split(':')[0]
        return nsp if nsp in nsids else name

    # TODO: Check for extension usages

    items = jadn.build_deps(schema)
    # out, roots = topo_sort(items)
    meta = schema['meta'] if 'meta' in schema else {}
    imports = meta['imports'] if 'imports' in meta else {}
    exports = meta['exports'] if 'exports' in meta else []
    defs = {i for i in items} | set(imports)
    refs = set().union([ns(r, imports) for i in items for r in items[i]]) | set(exports)
    oids = (OPTION_ID['enum'], OPTION_ID['pointer'])
    refs = {r[1:] if r[0] in oids else r for r in refs}         # Reference base type for derived enums/pointers
    return {
        'unreferenced': [str(k) for k in defs - refs],
        'undefined': [str(k) for k in refs - defs],
        'cycles': [],
    }


def loads(jadn_str):
    schema = json.loads(jadn_str)
    return check(schema)


def load(fname):
    with open(fname, encoding='utf-8') as f:
        schema = json.load(f)
    return check(schema)


def dumps(schema, level=0, indent=1, strip=False, nlevel=None):
    sp = level * indent * ' '
    sp2 = (level + 1) * indent * ' '
    sep2 = ',\n' if strip else ',\n\n'
    if isinstance(schema, dict):
        sep = ',\n' if level > 0 else sep2
        lines = []
        for k in schema:
            lines.append(sp2 + '"' + k + '": ' + dumps(schema[k], level + 1, indent, strip))
        return '{\n' + sep.join(lines) + '\n' + sp + '}'
    elif isinstance(schema, list):
        sep = ',\n' if level > 1 else sep2
        vals = []
        nest = schema and isinstance(schema[0], list)       # Not an empty list
        for v in schema:
            sp3 = sp2 if nest else ''
            vals.append(sp3 + dumps(v, level + 1, indent, strip, level))
        if nest:
            spn = (nlevel if nlevel else level) * indent * ' '
            return '[\n' + sep.join(vals) + '\n' + spn + ']'
        return '[' + ', '.join(vals) + ']'
    elif isinstance(schema, (numbers.Number, type(''))):
        return json.dumps(schema)
    return '???'


def dump(schema, fname, source='', strip=False):
    with open(fname, 'w') as f:
        if source:
            f.write('"Generated from ' + source + ', ' + datetime.ctime(datetime.now()) + '"\n\n')
        f.write(dumps(schema, strip=strip) + '\n')



from jadn.core import raise_error, analyze, check, dump, dumps, load, loads, data_dir
from jadn.utils import topts_s2d, ftopts_s2d, opts_d2s, get_optx, del_opt, cleanup_tagid, opts_sort, canonicalize
from jadn.utils import build_deps, get_config, jadn2typestr, typestr2jadn, jadn2fielddef, fielddef2jadn
import jadn.codec
import jadn.convert
import jadn.transform
import jadn.translate

__version__ = '0.5.9'